import { Program } from 'estree';
import Primitive from './Primitive';
import JsObject from './JsObject';
import { AsyncFunction } from './types';
export interface State {
    node: any;
    scope?: any;
    thisExpression?: any;
    done?: boolean;
    func_?: JsObject;
    funcThis_?: any;
    arguments?: any;
    doneArgs_?: boolean;
    doneExec_?: boolean;
    throwValue?: any;
    doneLeft?: boolean;
    doneRight?: boolean;
    value?: any;
    interpreter?: any;
    n?: any;
    array?: any;
    components?: boolean;
    leftSide?: any;
    doneGetter_?: any;
    leftValue?: any;
    doneCallee_?: any;
    n_?: any;
    label?: any;
    isLoop?: boolean;
    isSwitch?: boolean;
    doneSetter_?: any;
    doneObject?: any;
    doneBody?: any;
    member_?: any;
    isConstructor_?: any;
    doneVariable_?: any;
    doneObject_?: any;
    variable?: any;
    iterator?: any;
    object?: any;
    test?: any;
    doneLeft_?: any;
    doneRight_?: any;
    doneProperty_?: any;
    checked?: any;
    switchValue?: any;
    index?: any;
    valueToggle?: any;
    properties?: any;
    key?: any;
    kind?: any;
    argument?: any;
    doneBlock?: boolean;
    doneFinalizer?: any;
    mode?: any;
}
export interface InitFunc {
    (interpreter: Interpreter, scope: JsObject): any;
}
/**
 * @fileoverview Interpreting JavaScript in JavaScript.
 * @author fraser@google.com (Neil Fraser)
 */
export declare class Interpreter {
    UNDEFINED: Primitive;
    NULL: Primitive;
    NAN: Primitive;
    TRUE: Primitive;
    FALSE: Primitive;
    NUMBER_ZERO: Primitive;
    NUMBER_ONE: Primitive;
    STRING_EMPTY: Primitive;
    NUMBER: Primitive;
    BOOLEAN: Primitive;
    STRING: Primitive;
    OBJECT: any;
    FUNCTION: any;
    ARRAY: any;
    REGEXP: any;
    TYPE_ERROR: any;
    URI_ERROR: any;
    REFERENCE_ERROR: any;
    RANGE_ERROR: any;
    DATE: any;
    ERROR: any;
    EVAL_ERROR: any;
    SYNTAX_ERROR: any;
    ast: Program;
    initFunc_: InitFunc;
    paused_: boolean;
    polyfills_: string[];
    stateStack: State[];
    value: any;
    parentScope: any;
    /**
     * Create a new interpreter.
     * @param {string|!Object} code Raw JavaScript text or AST.
     * @param {Function=} opt_initFunc Optional initialization function.  Used to
     *     define APIs.  When called it is passed the interpreter object and the
     *     global scope object.
     * @constructor
     */
    constructor(code: any, opt_initFunc?: InitFunc);
    /**
     * Add more code to the interpreter.
     * @param {string|!Object} code Raw JavaScript text or AST.
     */
    appendCode(code: any): void;
    /**
     * Execute one step of the interpreter.
     * @return {boolean} True if a step was executed, false if no more instructions.
     */
    step(): boolean;
    /**
     * Execute the interpreter to program completion. Vulnerable to infinite loops.
     * @return {boolean} True if a execution is asynchonously blocked,
     *     false if no more instructions.
     */
    run(): boolean;
    /**
     * Initialize the global scope with buitin properties and functions.
     * @param {!JsObject} scope Global scope.
     */
    initGlobalScope(scope: JsObject): void;
    private createUriFunction(fn);
    /**
     * Initialize the Function class.
     * @param {!JsObject} scope Global scope.
     */
    initFunction(scope: any): void;
    /**
     * Initialize the Object class.
     * @param {!JsObject} scope Global scope.
     */
    initObject(scope: any): void;
    /**
     * Initialize the Array class.
     * @param {!JsObject} scope Global scope.
     */
    initArray(scope: any): void;
    /**
     * Initialize the Number class.
     * @param {!JsObject} scope Global scope.
     */
    initNumber(scope: any): void;
    /**
     * Initialize the String class.
     * @param {!JsObject} scope Global scope.
     */
    initString(scope: any): void;
    /**
     * Initialize the Boolean class.
     * @param {!JsObject} scope Global scope.
     */
    initBoolean(scope: any): void;
    /**
     * Initialize the Date class.
     * @param {!JsObject} scope Global scope.
     */
    initDate(scope: any): void;
    /**
     * Initialize Math object.
     * @param {!JsObject} scope Global scope.
     */
    initMath(scope: any): void;
    /**
     * Initialize Regular Expression object.
     * @param {!JsObject} scope Global scope.
     */
    initRegExp(scope: any): void;
    /**
     * Initialize JSON object.
     * @param {!JsObject} scope Global scope.
     */
    initJSON(scope: any): void;
    /**
     * Initialize the Error class.
     * @param {!JsObject} scope Global scope.
     */
    initError(scope: any): void;
    /**
     * Create a new data object for a primitive.
     * @param {number|string|boolean|null|undefined|RegExp} data Data to
     *     encapsulate.
     * @return {!Primitive|!JsObject} New data object.
     */
    createPrimitive(data: any): any;
    /**
     * Create a new data object.
     * @param {JsObject} parent Parent constructor function.
     * @return {!JsObject} New data object.
     */
    createObject(parent: any): JsObject;
    /**
     * Initialize a pseudo regular expression object based on a native regular
     * expression object.
     * @param {!JsObject} pseudoRegexp The existing object to set.
     * @param {!RegExp} nativeRegexp The native regular expression.
     * @return {!JsObject} Newly populated regular expression object.
     * @private
     */
    populateRegExp_(pseudoRegexp: any, nativeRegexp: any): any;
    /**
     * Create a new function.
     * @param {Object} node AST node defining the function.
     * @param {Object=} opt_scope Optional parent scope.
     * @return {!JsObject} New function.
     */
    createFunction(node: any, opt_scope?: any): JsObject;
    /**
     * Create a new native function.
     * @param {!Function} nativeFunc JavaScript function.
     * @return {!JsObject} New function.
     */
    createNativeFunction(nativeFunc: Function): JsObject;
    /**
     * Create a new native asynchronous function.
     * @param {!Function} asyncFunc JavaScript function.
     * @return {!JsObject} New function.
     */
    createAsyncFunction(asyncFunc: AsyncFunction): JsObject;
    /**
     * Converts from a native JS object or value to a JS interpreter object.
     * Can handle JSON-style values.
     * @param {*} nativeObj The native JS object to be converted.
     * @return {!JsObject|!Primitive} The equivalent
     *     JS interpreter object.
     */
    nativeToPseudo(nativeObj: any): any;
    /**
     * Converts from a JS interpreter object to native JS object.
     * Can handle JSON-style values.
     * @param {!JsObject|!Primitive} pseudoObj The JS
     *     interpreter object to be converted.
     * @return {*} The equivalent native JS object or value.
     */
    pseudoToNative(pseudoObj: any): any;
    /**
     * Fetch a property value from a data object.
     * @param {!JsObject|!Primitive} obj Data object.
     * @param {*} name Name of property.
     * @return {!JsObject|!Primitive|null} Property value
     *     (may be UNDEFINED), or null if an error was thrown and will be caught.
     */
    getProperty(obj: any, name: any): any;
    /**
     * Does the named property exist on a data object.
     * @param {!JsObject|!Primitive} obj Data object.
     * @param {*} name Name of property.
     * @return {boolean} True if property exists.
     */
    hasProperty(obj: any, name: any): boolean;
    /**
     * Set a property value on a data object.
     * @param {!JsObject} obj Data object.
     * @param {*} name Name of property.
     * @param {JsObject|Primitive} value
     *     New property value or null if getter/setter is described.
     * @param {Object=} opt_descriptor Optional descriptor object.
     * @return {!JsObject|undefined} Returns a setter function if one
     *     needs to be called, otherwise undefined.
     */
    setProperty(obj: any, name: any, value: any, opt_descriptor?: any): any;
    /**
     * Convenience method for adding a native function as a non-enumerable property
     * onto an object's prototype.
     * @param {!JsObject} obj Data object.
     * @param {*} name Name of property.
     * @param {!Function} wrapper Function object.
     */
    setNativeFunctionPrototype(obj: any, name: any, wrapper: any): void;
    /**
     * Delete a property value on a data object.
     * @param {!JsObject} obj Data object.
     * @param {*} name Name of property.
     * @return {boolean} True if deleted, false if undeletable.
     */
    deleteProperty(obj: any, name: any): boolean;
    /**
     * Returns the current scope from the stateStack.
     * @return {!JsObject} Current scope dictionary.
     */
    getScope(): any;
    /**
     * Create a new scope dictionary.
     * @param {!Object} node AST node defining the scope container
     *     (e.g. a function).
     * @param {JsObject} parentScope Scope to link to.
     * @return {!JsObject} New scope.
     */
    createScope(node: any, parentScope: any): JsObject;
    /**
     * Create a new special scope dictionary. Similar to createScope(), but
     * doesn't assume that the scope is for a function body. This is used for
     * the catch clause and with statement.
     * @param {!JsObject} parentScope Scope to link to.
     * @param {JsObject=} opt_scope Optional object to transform into
     *     scope.
     * @return {!JsObject} New scope.
     */
    createSpecialScope(parentScope: any, opt_scope?: any): any;
    /**
     * Retrieves a value from the scope chain.
     * @param {!JsObject|!Primitive} name Name of variable.
     * @return {!JsObject|!Primitive|null} The value
     *     or null if an error was thrown and will be caught.
     */
    getValueFromScope(name: any): any;
    /**
     * Sets a value to the current scope.
     * @param {!JsObject|!Primitive} name Name of variable.
     * @param {!JsObject|!Primitive} value Value.
     */
    setValueToScope(name: any, value: any): void;
    /**
     * Create a new scope for the given node.
     * @param {!Object} node AST node (program or function).
     * @param {!JsObject} scope Scope dictionary to populate.
     * @private
     */
    populateScope_(node: any, scope: any): void;
    /**
     * Gets a value from the scope chain or from an object property.
     * @param {!JsObject|!Primitive|!Array} left
     *     Name of variable or object/propname tuple.
     * @return {!JsObject|!Primitive|null} Value
     *     or null if an error was thrown and will be caught.
     */
    getValue(left: any): any;
    /**
     * Sets a value to the scope chain or to an object property.
     * @param {!JsObject|!Primitive|!Array} left
     *     Name of variable or object/propname tuple.
     * @param {!JsObject|!Primitive} value Value.
     * @return {!JsObject|undefined} Returns a setter function if one
     *     needs to be called, otherwise undefined.
     */
    setValue(left: any, value: any): any;
    /**
     * Throw an exception in the interpreter that can be handled by a
     * interpreter try/catch statement.  If unhandled, a real exception will
     * be thrown.  Can be called with either an error class and a message, or
     * with an actual object to be thrown.
     * @param {!JsObject} errorClass Type of error (if message is
     *   provided) or the value to throw (if no message).
     * @param {string=} opt_message Message being thrown.
     */
    throwException(errorClass: any, opt_message?: any): void;
    stepArrayExpression(): void;
    stepAssignmentExpression(): void;
    stepBinaryExpression(): void;
    stepBlockStatement(): void;
    stepBreakStatement(): void;
    stepCallExpression(): void;
    stepCatchClause(): void;
    stepConditionalExpression(): void;
    stepContinueStatement(): void;
    stepDoWhileStatement(): void;
    stepEmptyStatement(): void;
    stepEval_(): void;
    stepExpressionStatement(): void;
    stepForInStatement(): void;
    stepForStatement(): void;
    stepFunctionDeclaration(): void;
    stepFunctionExpression(): void;
    stepIdentifier(): void;
    stepIfStatement(): void;
    stepLabeledStatement(): void;
    stepLiteral(): void;
    stepLogicalExpression(): void;
    stepMemberExpression(): void;
    stepNewExpression(): void;
    stepObjectExpression(): void;
    stepProgram(): void;
    stepReturnStatement(): void;
    stepSequenceExpression(): void;
    stepSwitchStatement(): void;
    stepThisExpression(): void;
    stepThrowStatement(): void;
    stepTryStatement(): void;
    stepUnaryExpression(): void;
    stepUpdateExpression(): void;
    stepVariableDeclaration(): void;
    stepVariableDeclarator(): void;
    stepWithStatement(): void;
    stepWhileStatement(): void;
}
export default Interpreter;
